

import asyncio

from typing import Optional, Union, Any
from datetime import datetime, timedelta
from datetime import timezone

from aiogram.types import (
    CallbackQuery,
    Message,
    FSInputFile,
    InlineKeyboardMarkup,
    UNSET_PARSE_MODE,
)

from aiogram import Bot
from aiogram.enums.chat_type import ChatType

from usefulgram.enums import Const
from usefulgram.exceptions import MessageTooOld
from usefulgram.lazy.editor import MessageEditor
from usefulgram.lazy.sender import MessageSender
from usefulgram.lazy.stable_wait import StableWaiter
from usefulgram.lazy.callback_responder import CallbackAnswer


class LazyEditor:
    _callback: CallbackQuery
    _bot: Bot
    _stable: bool

    def __init__(self, callback: CallbackQuery, bot: Bot, stable: bool = False):
        self._callback = callback
        self._bot = bot
        self._stable = stable

    @staticmethod
    def _get_text_by_caption(
            text: Optional[str],
            caption: Optional[str]
    ) -> Optional[str]:

        if text is None or caption is not None:
            return caption

        return text

    @staticmethod
    def _get_message_text(text: Optional[str], message: Message) -> Optional[str]:
        if text is None:
            return message.text

        return text

    @staticmethod
    def _get_data_changes_status(
            message: Message,
            text: Optional[str],
            reply_markup: Optional[InlineKeyboardMarkup],
            video: Optional[FSInputFile],
            photo: Optional[FSInputFile],
    ) -> bool:

        media = photo or video

        if message.text != text:
            return True

        if message.reply_markup != reply_markup:
            return True

        if media:
            return True

        return False

    @staticmethod
    def _get_delta_between_current_and_message(
            message_date: datetime
    ) -> timedelta:

        current = datetime.now(tz=timezone.utc)

        return current - message_date

    @staticmethod
    def _get_time_allowed_edit_status(message: Message) -> bool:
        const_delta = timedelta(hours=Const.ALLOW_EDITING_DELTA)

        delta = LazyEditor._get_delta_between_current_and_message(
            message.date
        )

        return delta < const_delta

    @staticmethod
    def _get_message_in_chat_status(message: Message) -> bool:
        if message.chat.type == ChatType.CHANNEL:
            return True

        return False

    @staticmethod
    def _get_message_media_is_correct_status(
            message: Message,
            photo: Optional[FSInputFile],
            video: Optional[FSInputFile]
    ) -> bool:

        message_has_media = message.photo or message.video
        media = photo or video

        if media and not message_has_media:
            return False

        if not media and message_has_media:
            return False

        return True

    def _get_bot_allow_edit_status(
            self,
            message: Message,
            photo: Optional[FSInputFile],
            video: Optional[FSInputFile],
    ) -> bool:

        if self._get_message_in_chat_status(message):
            return True

        if not self._get_time_allowed_edit_status(message):
            return False

        if not self._get_message_media_is_correct_status(
            message=message, photo=photo, video=video
        ):
            return False

        return True

    @staticmethod
    async def _send_or_edit(
            bot: Bot,
            can_edit: bool,
            message: Message,
            text: Optional[str],
            photo: Optional[FSInputFile],
            video: Optional[FSInputFile],
            reply_markup: Optional[InlineKeyboardMarkup],
            parse_mode: Union[str],
            disable_web_page_preview: bool,
            on_onflict_do_nothing: bool
    ) -> Union[Message, bool]:

        if not can_edit:
            if on_onflict_do_nothing:
                return False

            text = LazyEditor._get_message_text(text, message=message)

            return await MessageSender.send(
                bot=bot,
                chat_id=message.chat.id,
                message_thread_id=message.message_thread_id,
                text=text,
                photo=photo,
                video=video,
                reply_markup=reply_markup,
                parse_mode=parse_mode,
                disable_web_page_preview=disable_web_page_preview
            )

        return await MessageEditor.edit(
            bot=bot,
            chat_id=message.chat.id,
            message_id=message.message_id,
            text=text,
            photo=photo,
            video=video,
            reply_markup=reply_markup,
            parse_mode=parse_mode,
            disable_web_page_preview=disable_web_page_preview
        )

    def _get_can_edit_status(
            self,
            message: Message,
            photo: Optional[FSInputFile],
            video: Optional[FSInputFile],
    ) -> bool:

        if message is None:
            return False

        if not self._get_bot_allow_edit_status(
                message=message,
                photo=photo,
                video=video
        ):
            return False

        return True

    @staticmethod
    def _get_stable_wait_time(message: Message) -> float:
        if message.edit_date:
            dt = datetime.fromtimestamp(message.edit_date, tz=timezone.utc)

        else:
            dt = message.date

        return StableWaiter.get_stable_wait_time(dt)

    async def edit(
            self,
            text: Optional[str] = None,
            photo: Optional[FSInputFile] = None,
            video: Optional[FSInputFile] = None,
            reply_markup: Optional[InlineKeyboardMarkup] = None,
            parse_mode: Union[str, Any] = UNSET_PARSE_MODE,
            disable_web_page_preview: bool = False,
            on_onflict_do_nothing: bool = False,
            answer_text: Optional[str] = None,
            answer_show_alert: bool = False,
            autoanswer: bool = True
    ) -> Union[Message, bool]:
        """
        Smart edit menager
        :param text: edit message text
        :param photo: edit message media
        :param video: edit message media
        :param reply_markup: edit message inline keyboard object
        :param parse_mode: Mode for parsing entities in the message text.
        See formatting options for more details.
        :param disable_web_page_preview:
        :param on_onflict_do_nothing:
        :param answer_text: if autoanser, it got callback.answer()
        :param answer_show_alert: if autoanser, it got callback.answer()
        :param autoanswer: should menager calls callback.answer()
        :return:
        """

        callback = self._callback
        message = callback.message

        if message is None:
            await callback.answer("Message too old")

            raise MessageTooOld

        if self._stable:
            await asyncio.sleep(self._get_stable_wait_time(message))

        if not self._get_data_changes_status(
                message=message,
                text=text,
                reply_markup=reply_markup,
                video=video,
                photo=photo,
        ):
            return message

        can_edit = self._get_can_edit_status(
            message=message,
            photo=photo,
            video=video,
        )

        result = await self._send_or_edit(
            bot=self._bot,
            can_edit=can_edit,
            message=message,
            text=text,
            photo=photo,
            video=video,
            reply_markup=reply_markup,
            parse_mode=parse_mode,
            disable_web_page_preview=disable_web_page_preview,
            on_onflict_do_nothing=on_onflict_do_nothing
        )

        await CallbackAnswer.auto_callback_answer(
            bot=self._bot,
            callback_id=callback.id,
            autoanswer=autoanswer,
            answer_text=answer_text,
            answer_show_alert=answer_show_alert
        )

        return result





from typing import Dict, Any, Awaitable, Callable, Optional, Union

from aiogram import BaseMiddleware, Bot
from aiogram.types import TelegramObject, CallbackQuery, Message, Update

from usefulgram.exceptions import BotIsUndefined
from usefulgram.parsing.decode import DecodeCallbackData
from usefulgram.lazy import LazyEditor
from usefulgram.lazy import LazySender


class StackerMiddleware(BaseMiddleware):
    def __init__(self, stable: bool = False, separator: str = "/"):
        self.separator = separator
        self.stable = stable

    async def __call__(
            self,
            handler: Callable[[TelegramObject, Dict[str, Any]], Awaitable[Any]],
            event: TelegramObject,
            data: Dict[str, Any]
    ):

        event_object: Union[CallbackQuery, Message]

        if not isinstance(event, (CallbackQuery, Message, Update)):
            raise await handler(event, data)

        if isinstance(event, Update):
            if event.message is not None:
                event_object = event.message

            elif event.callback_query is not None:
                event_object = event.callback_query

            else:
                raise await handler(event, data)

        else:
            event_object = event

        bot: Optional[Bot] = data.get("bot", None)

        if bot is None:
            raise BotIsUndefined

        data["sender"] = LazySender(
            event=event_object,
            bot=bot,
            stable=self.stable
        )

        if not isinstance(event_object, CallbackQuery):
            return await handler(event, data)

        data["decoder"] = DecodeCallbackData(
            callback_data=event_object.data,
            separator=self.separator
        )

        data["editor"] = LazyEditor(
            callback=event_object,
            bot=bot,
            stable=self.stable
        )

        return await handler(event, data)





"""
This code based on
https://github.com/wakaree/simple_echo_bot/blob/main/middlewares/throttling.py
"""

from typing import Callable, Dict, Any, Awaitable, MutableMapping, Optional

from aiogram import BaseMiddleware
from aiogram.types import TelegramObject, User

from cachetools import TTLCache

from usefulgram.exceptions import Throttling


class ThrottlingMiddleware(BaseMiddleware):
    RATE_LIMIT = 0.7
    SIMPLE = True

    def __init__(
            self, rate_limit: float = RATE_LIMIT, simple: bool = SIMPLE
    ) -> None:

        self._cache: MutableMapping[int, None] = TTLCache(
            maxsize=10_000, ttl=rate_limit
        )

        self.SIMPLE = simple

    async def __call__(
        self,
        handler: Callable[[TelegramObject, Dict[str, Any]], Awaitable[Any]],
        event: TelegramObject,
        data: Dict[str, Any],
    ) -> Optional[Any]:
        user: Optional[User] = data.get("event_from_user", None)

        if user is not None:
            if user.id in self._cache:
                self._cache[user.id] = None

                if self.SIMPLE:
                    return None

                raise Throttling()

            self._cache[user.id] = None

        return await handler(event, data)





from typing import Optional, Union
from datetime import date

from usefulgram.enums.calendar import CalendarEnum
from usefulgram.filters.parse_filters import BasePydanticFilter


class CalendarChangeButton(BasePydanticFilter):
    prefix: str = "calendar_change"
    year: Optional[int] = None
    month: Optional[int] = None


class CalendarInfoButton(BasePydanticFilter):
    prefix: str = "calendar_info"
    button_type: Union[CalendarEnum, str, None] = None


class CalendarDateFilter(BasePydanticFilter):
    date_value: Optional[date] = None



import asyncio

from typing import Union, Optional

from aiogram import Bot
from aiogram.types import (
    CallbackQuery,
    Message,
    FSInputFile,
    UNSET_PARSE_MODE,
    InlineKeyboardMarkup
)

from usefulgram.exceptions import MessageTooOld
from usefulgram.lazy.sender import MessageSender
from usefulgram.lazy.stable_wait import StableWaiter
from usefulgram.lazy.callback_responder import CallbackAnswer


class LazySender:
    _event: Union[CallbackQuery, Message]
    _bot: Bot
    _stable: bool

    def __init__(
            self,
            event: Union[CallbackQuery, Message],
            bot: Bot,
            stable: bool = False
    ):

        self._event = event
        self._bot = bot
        self._stable = stable

    @staticmethod
    def get_empty_text() -> str:
        spaces = " ⁣" * 65

        return f"⁣{spaces}⁣"

    async def send(
            self,
            text: Optional[str] = None,
            photo: Optional[FSInputFile] = None,
            video: Optional[FSInputFile] = None,
            reply_markup: Optional[InlineKeyboardMarkup] = None,
            parse_mode: Union[str] = UNSET_PARSE_MODE,
            disable_web_page_preview: bool = False,
            answer_text: Optional[str] = None,
            answer_show_alert: bool = False,
            autoanswer: bool = True
    ) -> Message:
        """
        Smart send menager

        :param text:
        :param photo:
        :param video:
        :param reply_markup:
        :param parse_mode:
        :param disable_web_page_preview:
        :param answer_text:
        :param answer_show_alert:
        :param autoanswer:
        :return:
        """

        if isinstance(self._event, CallbackQuery):
            if self._event.message is None:
                await self._event.answer("Message too old")

                raise MessageTooOld

            chat_id = self._event.message.chat.id
            thread_id = self._event.message.message_thread_id
            dt = self._event.message.date

        else:
            chat_id = self._event.chat.id
            thread_id = self._event.message_thread_id
            dt = self._event.date

        if self._stable:
            await asyncio.sleep(StableWaiter.get_stable_wait_time(dt))

        if text is None:
            text = self.get_empty_text()

        result = await MessageSender.send(
            bot=self._bot,
            chat_id=chat_id,
            text=text,
            message_thread_id=thread_id,
            photo=photo,
            video=video,
            reply_markup=reply_markup,
            parse_mode=parse_mode,
            disable_web_page_preview=disable_web_page_preview
        )

        if isinstance(self._event, CallbackQuery):
            await CallbackAnswer.auto_callback_answer(
                bot=self._bot,
                callback_id=self._event.id,
                autoanswer=autoanswer,
                answer_text=answer_text,
                answer_show_alert=answer_show_alert
            )

        return result





import calendar as python_calendar
from typing import Optional, Iterable, Coroutine, Any, Union
from datetime import datetime, date

from aiogram.types import InlineKeyboardMarkup, Message

from usefulgram.enums.calendar import CalendarEnum
from usefulgram.lazy import LazySender, LazyEditor
from usefulgram.keyboard import Builder, Row, Button
from usefulgram.utils.calendar_manager.filters import (
    CalendarChangeButton,
    CalendarInfoButton,
    CalendarDateFilter
)

from usefulgram.utils.calendar_manager.model import BaseCalendar, en_calendar


class Calendar:
    @staticmethod
    def _get_minus_month_and_year(month: int, year: int) -> tuple[int, int]:
        if month - 1 < 1:
            return 12, year - 1

        return month - 1, year

    @staticmethod
    def _get_plus_month_and_year(month: int, year: int) -> tuple[int, int]:
        if month + 1 > 12:
            return 1, year + 1

        return month + 1, year

    @staticmethod
    def _get_change_month_button(
            month: int,
            year: int,
            is_right: bool
    ) -> Button:

        if is_right:
            text = "⇨"

        else:
            text = "⇦"

        return Button(text, CalendarChangeButton(year=year, month=month))

    @staticmethod
    def _get_heading(weekdays: Iterable[str]) -> Row:
        heading_buttons = []

        for day in weekdays:
            heading_buttons.append(
                Button(
                    day, CalendarInfoButton(button_type=CalendarEnum.WEEKDAY))
            )

        return Row(*heading_buttons)

    @staticmethod
    def _get_format_date(month: int, year: int) -> str:
        return f"{month}.{year}"

    def _get_format_date_button(self, month: int, year: int) -> Button:
        text = self._get_format_date(month, year)

        return Button(text, CalendarInfoButton(button_type=CalendarEnum.DATE))

    @staticmethod
    def _get_current_date() -> date:
        return datetime.now().date()

    @staticmethod
    def _get_calendar(year: int, month: int) -> Iterable[date]:
        return python_calendar.Calendar().itermonthdates(year, month)

    @staticmethod
    def _get_calendars_day_buttons(
            days: Iterable[date],
            current_date: date,
            month: int,
            result_class: CalendarDateFilter
    ) -> list[Row]:

        row_list: list[Row] = []
        current_row: list[Button] = []

        for index, day in enumerate(days):
            if index % 7 == 0:
                row_list.append(Row(*current_row))
                current_row = []

            if day.month != month:
                current_row.append(
                    Button(
                        " ",
                        CalendarInfoButton(button_type=CalendarEnum.EMPTY))
                )

                continue

            if day == current_date:
                humanize_day = f"[{day.day}]"

            else:
                humanize_day = str(day.day)

            result_class.date_value = day

            current_row.append(
                Button(humanize_day, result_class)
            )

        row_list.append(Row(*current_row))

        return row_list

    def _get_days_buttons(
            self,
            month: int,
            year: int,
            result_class: CalendarDateFilter
    ) -> list[Row]:

        current_date = self._get_current_date()

        month_day = self._get_calendar(year, month)

        return self._get_calendars_day_buttons(
            days=month_day,
            current_date=current_date,
            month=month,
            result_class=result_class
        )

    def _get_bottom_buttons(self, month: int, year: int) -> Row:
        left_month, left_year = self._get_minus_month_and_year(month, year)
        right_month, right_year = self._get_plus_month_and_year(month, year)

        left_button = self._get_change_month_button(
            left_month, left_year, is_right=False
        )
        right_button = self._get_change_month_button(
            right_month, right_year, is_right=True
        )

        date_button = self._get_format_date_button(month, year)

        return Row(left_button, date_button, right_button)

    @staticmethod
    def _get_calendar_markup(
            heading: Row, day_buttons: list[Row], bottom_buttons: Row
    ) -> InlineKeyboardMarkup:

        return Builder(
            heading,
            *day_buttons,
            bottom_buttons
        )

    def show(
            self,
            sender: LazySender,
            month: int,
            year: int,
            result_class: CalendarDateFilter,
            text: Optional[str] = None,
            localization_class: BaseCalendar = en_calendar,
            editor: Optional[LazyEditor] = None
    ) -> Coroutine[Any, Any, Union[Message, bool]]:
        """
        Send or edit the calendar
        :param text:
        :param sender:
        :param month:
        :param year:
        :param result_class:
        :param localization_class:
        :param editor:
        :return:
        """

        heading = self._get_heading(localization_class.weekdays)
        day_buttons = self._get_days_buttons(month, year, result_class)
        bottom_buttons = self._get_bottom_buttons(month, year)

        markup = self._get_calendar_markup(
            heading, day_buttons, bottom_buttons
        )

        if editor is not None:
            return editor.edit(text=text, reply_markup=markup)

        return sender.send(text=text, reply_markup=markup)


calendar_manager = Calendar()





from typing import Final


class Const:
    DATETIME_FORMAT: Final[str] = "%Y%m%d%H%M%S"

    ALLOW_EDITING_DELTA: Final[int] = 47
    SECONDS_BETWEEN_OPERATION: Final[int] = 1

    STABLE_WAIT_TIME_SECONDS: Final[float] = 0.7





from typing import Optional, Any, Union

from aiogram.types import (
    InlineKeyboardButton,
    KeyboardButton,
    WebAppInfo,
    KeyboardButtonRequestChat,
    KeyboardButtonRequestUser,
    KeyboardButtonPollType,
    SwitchInlineQueryChosenChat,
    LoginUrl,
    CallbackGame
)

from usefulgram.parsing.encode import CallbackData


class Button(InlineKeyboardButton):
    def __init__(
            self,
            text: Union[str, int],
            *args: Any,
            prefix: Optional[str] = None,
            url: Optional[str] = None,
            callback_game: Optional[CallbackGame] = None,
            login_url: Optional[LoginUrl] = None,
            pay: Optional[bool] = None,
            switch_inline_query: Optional[str] = None,
            switch_inline_query_chosen_chat: Optional[
                SwitchInlineQueryChosenChat
            ] = None,
            switch_inline_query_current_chat: Optional[str] = None,
            separator: str = "/",
    ):
        prefix = self._get_prefix(prefix, args)

        if isinstance(text, int):
            text = str(text)

        if prefix is not None:
            callback_data = CallbackData(prefix, *args, separator=separator)

        else:
            callback_data = None

        super().__init__(
            text=text,
            url=url,
            callback_data=callback_data,
            callback_game=callback_game,
            login_url=login_url,
            pay=pay,
            switch_inline_query=switch_inline_query,
            switch_inline_query_chosen_chat=switch_inline_query_chosen_chat,
            switch_inline_query_current_chat=switch_inline_query_current_chat,
        )

    @staticmethod
    def _get_prefix(prefix: Optional[str], args: tuple[Any, ...]) -> Optional[str]:
        if prefix is not None:
            return prefix

        for obj in args:
            result = obj.__getattribute__("prefix")

            if result is not None:
                return result

        return None


class ReplyButton(KeyboardButton):
    def __init__(
            self,
            text: str,
            web_app: Optional[WebAppInfo] = None,
            request_chat: Optional[KeyboardButtonRequestChat] = None,
            request_user: Optional[KeyboardButtonRequestUser] = None,
            request_contact: Optional[bool] = None,
            request_poll: Optional[KeyboardButtonPollType] = None
    ):
        super().__init__(
            text=text,
            web_app=web_app,
            request_chat=request_chat,
            request_user=request_user,
            request_contact=request_contact,
            request_poll=request_poll
        )





NoOneButtonParamIsFilled = ValueError("No one button parameters is filled")

DifferentButtonsInMatrix = ValueError("Only one type of button is allowed in the markup")

UnknownButtonType = ValueError("UnknownButtonType")

TooMoreCharacters = ValueError(
    "The callback data can be only 64 bytes"
    "(~62 or less characters because separator)"
)

RecursionObjectParse = ValueError("Now objects cannot contain objects")

WrongObjectType = ValueError("Wrong object type in the decode")

CantEditMedia = ValueError("Error for edit message media")

BotIsUndefined = ValueError("Bot is not defined in StakerMiddleware")

MessageTooOld = ValueError("Message too old and has not message param")

MessageTextIsNone = ValueError("Message text is None")

UndefinedMagicFilterModel = ValueError("Magic filter model is undefined")

UndefinedType = ValueError("Error event type in a middleware")

CallbackEventWasNotGiven = ValueError("Callback event was not given to decoder")


class Throttling(Exception):
    def __init__(self):
        self.message = "User got trottling exception"

        super().__init__(self.message)





from typing import Optional, Union, Coroutine, Any

from aiogram.types import (
    Message,
    InputMediaPhoto,
    InputMediaVideo,
    FSInputFile,
    InlineKeyboardMarkup,
    UNSET_PARSE_MODE,
)

from aiogram import Bot

from usefulgram.exceptions import CantEditMedia


class MessageEditor:
    @staticmethod
    def edit(
            bot: Bot,
            chat_id: int,
            message_id: int,
            text: Optional[str] = None,
            photo: Optional[FSInputFile] = None,
            video: Optional[FSInputFile] = None,
            reply_markup: Optional[InlineKeyboardMarkup] = None,
            parse_mode: Union[str] = UNSET_PARSE_MODE,
            disable_web_page_preview: bool = False
    ) -> Coroutine[Any, Any, Union[Message, bool]]:

        if photo or video:
            media: Union[InputMediaVideo, InputMediaPhoto]

            if photo:
                media = InputMediaPhoto(
                    media=photo,
                    caption=text,
                    parse_mode=parse_mode
                )

            elif video:
                media = InputMediaVideo(
                    media=video,
                    caption=text,
                    parse_mode=parse_mode
                )

            else:
                raise CantEditMedia

            return bot.edit_message_media(
                chat_id=chat_id,
                message_id=message_id,
                media=media,
                reply_markup=reply_markup,
            )

        if text is None:
            return bot.edit_message_reply_markup(
                chat_id=chat_id,
                message_id=message_id,
                reply_markup=reply_markup,
            )

        return bot.edit_message_text(
            chat_id=chat_id,
            message_id=message_id,
            text=text,
            reply_markup=reply_markup,
            parse_mode=parse_mode,
            disable_web_page_preview=disable_web_page_preview
        )





import sys

from typing import Any

from enum import Enum
from decimal import Decimal

from datetime import date, time, datetime

from pydantic import BaseModel

from usefulgram.exceptions import TooMoreCharacters
from usefulgram.enums import Const


class _Additional:
    @staticmethod
    def _add_separator(item: str, separator: str = "&") -> str:
        return f"{item}{separator}"

    def _object_to_str(self, item: object) -> str:
        if isinstance(item, BaseModel):
            fields = item.model_fields

        else:
            fields = item.__annotations__

        annotations_keys = list(fields.keys())

        result = ""

        for key in annotations_keys:
            if key == "prefix":
                continue

            values = item.__getattribute__(f"{key}")

            string_value = self._to_str(values, is_recursion=True)

            result += self._add_separator(string_value)

        return result[:-1]

    def _to_str(self, item: Any, is_recursion: bool = False) -> str:
        if item is None:
            return ""

        if isinstance(item, bool):
            return str(int(item))

        if isinstance(item, str):
            return item

        if isinstance(item, (int, float, Decimal)):
            return str(item)

        if isinstance(item, (datetime, date, time)):
            return item.strftime(Const.DATETIME_FORMAT)

        if isinstance(item, Enum):
            return item.name

        if is_recursion:
            return self._object_to_str(item)

        try:
            return self._object_to_str(item)

        except AttributeError:
            return f"{item}"

    def __call__(self, *args: Any) -> str:
        if args == ():
            return ""

        result = ""

        for item in args:
            result += self._add_separator(self._to_str(item))

        result = result[:-1]

        return result


class _CallbackData:
    @staticmethod
    def _get_str_callback_data(
            prefix: str, additional: str, separator: str
    ) -> str:

        return f"{prefix}{separator}{additional}"

    @staticmethod
    def _check_callback_data_bytes(callback_data: str) -> bool:
        size = sys.getsizeof(callback_data)

        true_size = size - 37  # 37 - is a system empty string lenght

        if true_size < 64:
            return True

        raise TooMoreCharacters

    def __call__(
            self, prefix: str,
            *args: Any,
            separator: str = "/") -> str:

        additional = AdditionalInstance(*args)

        callback_data = self._get_str_callback_data(
            prefix, additional, separator
        )

        self._check_callback_data_bytes(callback_data)

        return callback_data


AdditionalInstance = _Additional()
CallbackData = _CallbackData()





from datetime import datetime, timedelta, timezone

from usefulgram.enums.const import Const


class StableWaiter:
    @staticmethod
    def _get_delta_between_current_and_datetime(
            message_datetime: datetime
    ) -> timedelta:

        current = datetime.now(tz=timezone.utc)

        return current - message_datetime

    @staticmethod
    def get_stable_wait_time(dt: datetime) -> float:
        delta = StableWaiter._get_delta_between_current_and_datetime(dt)

        total_seconds = delta.total_seconds()

        if total_seconds < 0:
            total_seconds = 0

        wait_time = Const.STABLE_WAIT_TIME_SECONDS - total_seconds

        return round(wait_time, 3)





import asyncio
from typing import Optional

from aiogram import Bot

from usefulgram.enums.const import Const


class CallbackAnswer:
    @staticmethod
    async def auto_callback_answer(
            bot: Bot,
            callback_id: str,
            autoanswer: bool = True,
            answer_text: Optional[str] = None,
            answer_show_alert: bool = False
    ) -> Optional[bool]:

        if not autoanswer:
            return None

        await asyncio.sleep(Const.SECONDS_BETWEEN_OPERATION)

        return await bot.answer_callback_query(
            callback_query_id=callback_id,
            text=answer_text,
            show_alert=answer_show_alert
        )





from . import lazy
from . import keyboard
from . import filters
from . import middlewares
from . import exceptions
from . import enums


__all__ = (
    "lazy",
    "keyboard",
    "filters",
    "middlewares",
    "exceptions"
)





from typing import Union, Any, Optional, Type
from datetime import datetime, date, time

from usefulgram.enums import Const
from usefulgram.exceptions import WrongObjectType

from enum import Enum

from pydantic import BaseModel

from contextlib import suppress


class DecodeCallbackData:
    prefix: str
    additional: list[str]

    @staticmethod
    def _get_prefix_and_additional(
            callback_data: str,
            separator: str
    ) -> tuple[str, list[str]]:

        split_data = callback_data.split(separator)

        additional = split_data[1].split("&")

        return split_data[0], additional

    @staticmethod
    def _get_empty_prefix_and_additional() -> tuple[str, list[str]]:
        return "", []

    def __init__(self, callback_data: Optional[str], separator: str = "/"):
        if callback_data is not None:
            self.prefix, self.additional = self._get_prefix_and_additional(
                callback_data, separator
            )

            return

        self.prefix, self.additional = self._get_empty_prefix_and_additional()

    def _convert_typing_object_to_type(
            self, obj_value: str, *args: type
    ) -> Any:

        for obj_type in args:
            with suppress(ValueError, TypeError):
                return self._convert_str_to_type(
                    obj_value=obj_value,
                    obj_type=obj_type
                )

        raise WrongObjectType

    def _convert_str_to_type(
            self,
            obj_value: str,
            obj_type: Any
    ) -> Any:

        if obj_value == "":
            return None

        # Optional[...] or Union[...] checker
        if not isinstance(obj_type, type):
            irregular_types = obj_type.__getattribute__("__args__")

            return self._convert_typing_object_to_type(
                obj_value, *irregular_types
            )

        if issubclass(obj_type, bool):
            return bool(int(obj_value))

        if issubclass(obj_type, datetime):
            return datetime.strptime(obj_value, Const.DATETIME_FORMAT)

        if issubclass(obj_type, date):
            return datetime.strptime(obj_value, Const.DATETIME_FORMAT).date()

        if issubclass(obj_type, time):
            return datetime.strptime(obj_value, Const.DATETIME_FORMAT).time()

        if issubclass(obj_type, Enum):
            return obj_type[obj_value]

        try:
            return obj_type(obj_value)  # type: ignore

        except (ValueError, AttributeError):
            raise WrongObjectType

    @staticmethod
    def _is_dataclass(obj_type: Any) -> bool:
        if isinstance(obj_type, type):
            return issubclass(obj_type, BaseModel)

        try:
            irregular_types = obj_type.__getattribute__("__args__")

        except (ValueError, AttributeError):
            return False

        for irregular_type in irregular_types:
            if issubclass(irregular_type, BaseModel):
                return True

        return False

    @staticmethod
    def _get_dataclass_obj(obj_type: Any) -> Type[BaseModel]:
        if isinstance(obj_type, type):
            if issubclass(obj_type, BaseModel):
                return obj_type

        irregular_types = obj_type.__getattribute__("__args__")

        for irregular_type in irregular_types:
            if issubclass(irregular_type, BaseModel):
                return irregular_type

        raise WrongObjectType

    def _iter_key_and_type(
            self, keys: list[str], objects_types: list[type],
            add_prefix: bool, start_additional_value: int = 0
    ) -> dict[str, Any]:

        return_param: dict[str, Any] = {}

        if add_prefix:
            return_param["prefix"] = self.prefix

        additional_value = start_additional_value

        for key, obj_type in zip(keys, objects_types):

            if key == "prefix":
                return_param[key] = self._convert_str_to_type(
                    self.prefix, obj_type
                )

                continue

            if self._is_dataclass(obj_type):
                dataclass_obj = self._get_dataclass_obj(obj_type)

                keys, values = self._get_key_and_values_by_obj(
                    format_object=dataclass_obj
                )

                params = self._iter_key_and_type(
                    keys=keys,
                    objects_types=values,
                    add_prefix=False,
                    start_additional_value=additional_value
                )

                additional_value += len(keys)

                return_param[key] = self._get_convert_obj(dataclass_obj, params)

                continue

            return_param[key] = self._convert_str_to_type(
                self.additional[additional_value], obj_type
            )

            additional_value += 1

        return return_param

    @staticmethod
    def _get_convert_obj(
            format_obj: type, params: dict[str, Any]
    ) -> Union[BaseModel, object]:
        return format_obj(**params)

    @staticmethod
    def _get_key_and_values_by_obj(
            format_object: type
    ) -> tuple[list[str], list[Any]]:

        if issubclass(format_object, BaseModel):
            fields = format_object.model_fields
            values = [i.annotation for i in fields.values()]

            return list(fields.keys()), values

        fields = format_object.__annotations__

        another_values = list(fields.values())

        return list(fields.keys()), another_values

    def to_format(
            self, format_object: type, add_prefix: bool = False
    ) -> Union[BaseModel, object]:

        keys, values = self._get_key_and_values_by_obj(format_object=format_object)

        obj_params = self._iter_key_and_type(keys, values, add_prefix)

        return self._get_convert_obj(format_object, obj_params)

    @staticmethod
    def class_to_dict(class_: Union[BaseModel, object]) -> dict[str, Any]:
        result_dict = {"prefix": class_.__getattribute__("prefix")}

        if isinstance(class_, BaseModel):
            fields = class_.model_fields

        else:
            fields = class_.__annotations__

        for key in fields.keys():
            result_dict[key] = class_.__getattribute__(key)

        return result_dict





from typing import Optional, Union, Coroutine, Any

from aiogram.types import (
    Message,
    FSInputFile,
    InlineKeyboardMarkup,
    UNSET_PARSE_MODE,
)

from aiogram import Bot

from usefulgram.exceptions import MessageTextIsNone


class MessageSender:
    @staticmethod
    def send(
            bot: Bot,
            chat_id: int,
            text: Optional[str],
            message_thread_id: Optional[int] = None,
            photo: Optional[FSInputFile] = None,
            video: Optional[FSInputFile] = None,
            reply_markup: Optional[InlineKeyboardMarkup] = None,
            parse_mode: Union[str] = UNSET_PARSE_MODE,
            disable_web_page_preview: bool = False,
    ) -> Coroutine[Any, Any, Message]:

        if photo is not None:
            return bot.send_photo(
                chat_id=chat_id,
                message_thread_id=message_thread_id,
                photo=photo,
                caption=text,
                reply_markup=reply_markup,
                parse_mode=parse_mode,
            )

        if video is not None:
            return bot.send_video(
                chat_id=chat_id,
                message_thread_id=message_thread_id,
                video=video,
                caption=text,
                reply_markup=reply_markup,
                parse_mode=parse_mode,
            )

        if text is None:
            raise MessageTextIsNone

        return bot.send_message(
            chat_id=chat_id,
            message_thread_id=message_thread_id,
            text=text,
            reply_markup=reply_markup,
            parse_mode=parse_mode,
            disable_web_page_preview=disable_web_page_preview
        )





from aiogram.types import (
    InlineKeyboardButton,
    KeyboardButton,
)


class Row:
    rows: list[InlineKeyboardButton]

    def __init__(self, *args: InlineKeyboardButton):
        self.rows = [*args]

    def get_rows(self) -> list[InlineKeyboardButton]:
        return self.rows


class ReplyRow:
    rows: list[KeyboardButton]

    def __init__(self, *args: KeyboardButton):
        self.rows = [*args]

    def get_rows(self) -> list[KeyboardButton]:
        return self.rows





from pydantic import BaseModel


# this.. temporary solution

class BaseCalendar(BaseModel):
    weekdays: tuple[str, str, str, str, str, str, str]


en_calendar = BaseCalendar(
    weekdays=(
        "Mo", "Tu", "We", "Th", "Fr", "Sa", "Su"
    )
)

ru_calendar = BaseCalendar(
    weekdays=(
        "Пн", "Вт", "Ср", "Чт", "Пт", "Сб", "Вс"
    )
)





from .exceptions import (
    DifferentButtonsInMatrix,
    UnknownButtonType,
    NoOneButtonParamIsFilled,
    TooMoreCharacters,
    RecursionObjectParse,
    WrongObjectType,
    CantEditMedia,
    BotIsUndefined,
    MessageTooOld,
    MessageTextIsNone,
    UndefinedMagicFilterModel,
    UndefinedType,
    CallbackEventWasNotGiven,
    Throttling
)





from typing import Optional, Union, Dict, Any, Tuple, Iterable
from abc import ABC, abstractmethod

from pydantic import BaseModel

from aiogram.filters import BaseFilter
from aiogram.types import CallbackQuery

from aiogram.filters.magic_data import MagicData
from aiogram.utils.magic_filter import MagicFilter

from usefulgram.parsing.decode import DecodeCallbackData
from usefulgram.exceptions import UndefinedMagicFilterModel


class _BaseMagicFilter(ABC):
    _magic_filter: Optional[tuple[MagicFilter]] = None

    def filter(self, *args: MagicFilter):
        self.__setattr__("_magic_filter", args)

        return self

    @staticmethod
    async def _ckeck_magic_filter(
            callback: CallbackQuery,
            magic_filter: tuple[MagicFilter],
            values: Optional[tuple[Any, ...]] = None,
            model: Optional[dict[Any, Any]] = None
    ) -> bool:

        for filter_obj in magic_filter:
            filter_instance = MagicData(filter_obj)

            if values is not None:
                result = await filter_instance(callback, *values)

            elif model is not None:
                result = await filter_instance(callback, **model)

            else:
                raise UndefinedMagicFilterModel

            if not result:
                return False

        return True

    def get_magic_filter(self) -> Any:
        try:
            return self.__getattribute__("_magic_filter")

        except AttributeError:
            return None


class _BaseDataclassesFilter(_BaseMagicFilter, ABC):
    @abstractmethod
    async def __call__(self, callback: CallbackQuery, decoder: DecodeCallbackData):
        """
        This abstract method should use the get_filter method
        :param callback:
        :param decoder:
        :return:
        """
        pass

    async def get_filter(
            self,
            callback: CallbackQuery,
            decoder: DecodeCallbackData,
            fields_name: Iterable[str]
    ) -> Union[bool, Dict[str, Any]]:

        try:
            callback_data_model = decoder.to_format(type(self), add_prefix=True)

        except (AttributeError, ValueError, IndexError, KeyError):
            return False

        for item_name in fields_name:
            if item_name == "_magic_filter":
                continue

            filter_value = self.__getattribute__(item_name)

            if filter_value is None:
                continue

            data_value = callback_data_model.__getattribute__(item_name)

            if data_value != filter_value:
                return False

        magic_filter: Optional[tuple[MagicFilter]] = super().get_magic_filter()

        if magic_filter is not None:
            if await self._ckeck_magic_filter(
                callback=callback,
                magic_filter=magic_filter,
                model=decoder.class_to_dict(callback_data_model)
            ):
                return decoder.class_to_dict(callback_data_model)

            return False

        return decoder.class_to_dict(callback_data_model)


class BasePydanticFilter(_BaseDataclassesFilter, BaseModel, BaseFilter):
    prefix: Optional[str] = None

    async def __call__(
            self,
            callback: CallbackQuery,
            decoder: DecodeCallbackData
    ) -> Union[bool, Dict[str, Any]]:

        fields = self.model_fields.keys()

        return await self.get_filter(
            callback=callback,
            decoder=decoder,
            fields_name=fields
        )


class CallbackPrefixFilter(BaseFilter):
    def __init__(self, prefix: str):
        """

        :param prefix: first argument in callback data
        """
        self.prefix = prefix

    async def __call__(self, message: CallbackQuery,
                       decoder: DecodeCallbackData) -> bool:

        if self.prefix == decoder.prefix:
            return True

        return False


class ItarationFilter(_BaseMagicFilter):
    def __init__(self, item_number: int = 0):
        self._item_number: int = item_number
        self._operation: Optional[Tuple[str, Any]] = None
        self._error_result: bool = False

    def __getitem__(self, item: int):
        self._item_number = item

        return self

    @staticmethod
    def _convert(other: str) -> Optional[int]:
        try:
            return int(other)

        except ValueError:
            return None

    def __eq__(self, other: Any):
        self._operation = ("eq", other)

        return self

    def __ne__(self, other: Any):
        self._operation = ("ne", other)

        return self

    def __lt__(self, other: Any):
        self._operation = ("lt", other)

        return self

    def __gt__(self, other: Any):
        self._operation = ("gt", other)

        return self

    def __le__(self, other: Any):
        self._operation = ("le", other)

        return self

    def __ge__(self, other: Any):
        self._operation = ("ge", other)

        return self

    @staticmethod
    def _do_operation(operation: str, first_item: Any, second_item: Any):
        if operation == "eq":
            return first_item == second_item

        if operation == "ne":
            return first_item != second_item

        first_item = ItarationFilter._convert(first_item)

        if first_item is False:
            return False

        if operation == "lt":
            return first_item < second_item

        if operation == "gt":
            return first_item > second_item

        if operation == "le":
            return first_item <= second_item

        if operation == "ge":
            return first_item >= second_item

    async def __call__(
            self, callback: CallbackQuery,
            decoder: DecodeCallbackData
    ):

        magic_filter: Optional[tuple[MagicFilter]] = super().get_magic_filter()

        if self._error_result:
            return False

        values = (decoder.prefix, *decoder.additional)

        first_item = values[self._item_number]

        if magic_filter is not None:
            if await self._ckeck_magic_filter(
                    callback=callback,
                    magic_filter=magic_filter,
                    values=values
            ):
                return True

            return False

        if self._operation is None:
            return False

        operation, second_item = self._operation

        return self._do_operation(operation, first_item, second_item)




from typing import Optional, Any, TypeVar

from aiogram.types import (
    InlineKeyboardButton,
    InlineKeyboardMarkup,
    KeyboardButton,
    ReplyKeyboardMarkup,
)

from usefulgram.keyboard.rows import Row, ReplyRow


T = TypeVar("T")


class BaseBuilder:
    @staticmethod
    def _get_adjust(rows: list[list[T]], adjust: int) -> list[list[T]]:
        buttons: list[T] = []

        for button_row in rows:
            for button in button_row:
                buttons.append(button)

        rows_len = len(buttons)

        result: list[list[T]] = []
        row: list[T] = []

        for index in range(0, rows_len):
            row.append(buttons[index])

            if (index + 1) % adjust == 0:
                result.append(row.copy())

                row.clear()

        if row:
            result.append(row.copy())

        return result

    @staticmethod
    def _get_inline_keyboard(
            rows: tuple[Row, ...],
            adjust: Optional[int]
    ) -> list[list[InlineKeyboardButton]]:

        rows_list = []

        for row in rows:
            rows_list.append(row.get_rows())

        return BaseBuilder._get_different_keyboar(rows_list, adjust)

    @staticmethod
    def _get_reply_keyboard(
            rows: tuple[ReplyRow, ...],
            adjust: Optional[int]
    ) -> list[list[KeyboardButton]]:

        rows_list = []

        for row in rows:
            rows_list.append(row.get_rows())

        return BaseBuilder._get_different_keyboar(rows_list, adjust)

    @staticmethod
    def _get_different_keyboar(
            rows: list[list[T]],
            adjust: Optional[int]
    ) -> list[list[T]]:
        if not rows:
            return []

        if adjust is None or adjust < 0:
            return rows

        return BaseBuilder._get_adjust(rows, adjust)

    @staticmethod
    def ignore_params(*args) -> tuple[Any, ...]:
        return args


class Builder(BaseBuilder, InlineKeyboardMarkup):
    def __new__(
            cls,
            *rows: Row,
            adjust: Optional[int] = None,
    ):
        keyboard = cls._get_inline_keyboard(rows, adjust)

        return InlineKeyboardMarkup(inline_keyboard=keyboard)

    def __init__(self, *rows: Row, adjust: Optional[int] = None):
        # This code is never get and made for Pydantic plugin
        super().__init__(inline_keyboard=[])

        self.ignore_params(rows, adjust)


class ReplyBuilder(BaseBuilder, ReplyKeyboardMarkup):
    def __new__(
            cls,
            *rows: ReplyRow,
            adjust: Optional[int] = None,
            resize_keayboard: Optional[bool] = None,
            is_persistent: Optional[bool] = None,
            one_time_keyboard: Optional[bool] = None,
    ):
        keyboard = cls._get_reply_keyboard(rows, adjust)

        return ReplyKeyboardMarkup(
            keyboard=keyboard,
            resize_keyboard=resize_keayboard,
            is_persistent=is_persistent,
            one_time_keyboard=one_time_keyboard
        )

    def __init__(
            self,
            *rows: ReplyRow,
            adjust: Optional[int] = None,
            resize_keayboard: Optional[bool] = None,
            is_persistent: Optional[bool] = None,
            one_time_keyboard: Optional[bool] = None,
    ):
        # This code is never get and made for Pydantic plugin

        super().__init__(keyboard=[])

        self.ignore_params(
            rows,
            adjust,
            resize_keayboard,
            is_persistent,
            one_time_keyboard
        )
